@@ -1,6 +1,7 @@
 /*
- * Copyright (C) 2007 daniel g. siegel <dgsiegel@gmail.com>
- * Copyright (C) 2007 Jaap Haitsma <jaap@haitsma.org>
+ * Copyright (C) 2007,2008 daniel g. siegel <dgsiegel@gmail.com>
+ * Copyright (C) 2007,2008 Jaap Haitsma <jaap@haitsma.org>
+ * Copyright (C) 2008 Patryk Zawadzki <patrys@pld-linux.org>
  *
  * Licensed under the GNU General Public License Version 2
  *
@@ -19,7 +20,7 @@
  */
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include <cheese-config.h>
 #endif
 
 #include <string.h>
@@ -28,23 +29,25 @@
 #include <gdk/gdkkeysyms.h>
 #include <glib.h>
 #include <glib/gi18n.h>
+#include <gio/gio.h>
 
 #include <gst/interfaces/xoverlay.h>
 #include <gtk/gtk.h>
-#include <libgnomevfs/gnome-vfs.h>
 #include <libebook/e-book.h>
 #include <glade/glade.h>
 
+#include "cheese-countdown.h"
 #include "cheese-effect-chooser.h"
 #include "cheese-fileutil.h"
 #include "cheese-gconf.h"
 #include "cheese-thumb-view.h"
 #include "cheese-window.h"
-#include "cheese-audio-play.h"
 #include "ephy-spinner.h"
+#include "gst-audio-play.h"
 
 #define GLADE_FILE PACKAGE_DATADIR"/cheese.glade"
 #define UI_FILE PACKAGE_DATADIR"/cheese-ui.xml"
+#define SHUTTER_SOUNDS 5
 
 typedef enum
 {
@@ -65,10 +68,10 @@ typedef struct
   CheeseWebcam *webcam;
   WebcamMode webcam_mode;
   CheeseGConf *gconf;
-  CheeseAudioPlay *audio_play;
 
   GtkWidget *window;
   GtkWidget *notebook;
+  GtkWidget *notebook_bar;
 
   GtkWidget *main_vbox;
 
@@ -76,6 +79,8 @@ typedef struct
   GtkWidget *effect_chooser;
   GtkWidget *throbber_frame;
   GtkWidget *throbber;
+  GtkWidget *countdown_frame;
+  GtkWidget *countdown;
 
   GtkWidget *button_effects;
   GtkWidget *button_photo;
@@ -107,27 +112,84 @@ typedef struct
 
   GtkUIManager *ui_manager;
 
+  int audio_play_counter;
+  GRand *rand;
+
 } CheeseWindow;
 
+static gboolean
+cheese_open_uri (GtkWindow *parent, const char *url, GError **error)
+{
+  gboolean ret;
+  char *cmdline;
+  GdkScreen *screen;
+
+  screen = gtk_window_get_screen (GTK_WINDOW (parent));
+  cmdline = g_strconcat ("xdg-open ", url, NULL);
+  ret = gdk_spawn_command_line_on_screen (screen, cmdline, error);
+  g_free (cmdline);
+
+  return ret;
+}
 
 /* Make url in about dialog clickable */
 static void
-cheese_about_dialog_handle_url (GtkAboutDialog *about, const char *link, gpointer data)
+cheese_about_dialog_handle_url (GtkAboutDialog *dialog, const char *url, gpointer data)
 {
-  gnome_vfs_url_show (link);
+  GError *error = NULL;
+  GtkWidget *error_dialog;
+  gboolean ret;
+
+  ret = cheese_open_uri (GTK_WINDOW (dialog), url, &error);
+  if (ret == FALSE) {
+    error_dialog = gtk_message_dialog_new (GTK_WINDOW (dialog), 
+                                           GTK_DIALOG_DESTROY_WITH_PARENT, 
+		 	                   GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, 
+				           _("Failed to open browser to show:\n%s"), url); 
+    gtk_dialog_run (GTK_DIALOG (error_dialog));
+    gtk_widget_destroy (error_dialog);
+    g_error_free(error);
+  }
 }
 
 /* Make email in about dialog clickable */
 static void
-cheese_about_dialog_handle_email (GtkAboutDialog *about, const char *link, gpointer data)
+cheese_about_dialog_handle_email (GtkAboutDialog *dialog, const char *email, gpointer data)
 {
-  char *address;
+  char *uri;
+  GError *error = NULL;
+  GtkWidget *error_dialog;
+  gboolean ret;
 
-  address = g_strdup_printf ("mailto:%s", link);
-  gnome_vfs_url_show (address);
-  g_free (address);
+  uri = g_strconcat ("mailto:", email, NULL);
+
+  ret = cheese_open_uri (GTK_WINDOW (dialog), uri, &error);
+  if (ret == FALSE) {
+    error_dialog = gtk_message_dialog_new (GTK_WINDOW (dialog), 
+                                           GTK_DIALOG_DESTROY_WITH_PARENT, 
+		 	                   GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, 
+				           _("Failed to open email client to send message to:\n%s"), email); 
+    gtk_dialog_run (GTK_DIALOG (error_dialog));
+    gtk_widget_destroy (error_dialog);
+    g_error_free(error);
+  }
+  g_free (uri);
 }
 
+static char *
+audio_play_get_filename (CheeseWindow *cheese_window)
+{
+  char *filename;
+  if (cheese_window->audio_play_counter > 21)
+   filename = g_strdup_printf ("%s/sounds/shutter%i.ogg", PACKAGE_DATADIR,
+                               g_rand_int_range (cheese_window->rand, 1, SHUTTER_SOUNDS));
+  else
+   filename = g_strdup_printf ("%s/sounds/shutter0.ogg", PACKAGE_DATADIR);
+
+  cheese_window->audio_play_counter++;
+
+  return filename;
+}
 
 /* standard event handler */
 static int
@@ -155,14 +217,25 @@ cheese_window_video_saved_cb (CheeseWebcam *webcam, CheeseWindow *cheese_window)
   gtk_widget_set_sensitive (cheese_window->take_picture, TRUE);
 }
 
+
 static void
 cheese_window_cmd_close (GtkWidget *widget, CheeseWindow *cheese_window)
 {
-  g_object_unref (cheese_window->audio_play);
+  g_free (cheese_window->rand);
   g_object_unref (cheese_window->webcam);
   g_object_unref (cheese_window->actions_main);
   g_object_unref (cheese_window->actions_photo);
+  g_object_unref (cheese_window->actions_mail);
+  g_object_unref (cheese_window->actions_toggle);
+  g_object_unref (cheese_window->actions_effects);
+  g_object_unref (cheese_window->actions_file);
+  g_object_unref (cheese_window->actions_video);
+  g_object_unref (cheese_window->actions_account_photo);
+  g_object_unref (cheese_window->actions_fspot);
+  g_object_unref (cheese_window->actions_flickr);
   g_object_unref (cheese_window->ui_manager);
+  g_object_unref (cheese_window->gconf);
+
   g_free (cheese_window);
   gtk_main_quit ();
 }
@@ -172,13 +245,25 @@ cheese_window_cmd_open (GtkWidget *widget, CheeseWindow *cheese_window)
 {
   char *uri;
   char *filename;
+  gboolean ret;
+  GError *error = NULL;
+  GtkWidget *dialog;
 
   filename = cheese_thumb_view_get_selected_image (CHEESE_THUMB_VIEW (cheese_window->thumb_view));
   g_return_if_fail (filename);
-
-  uri = gnome_vfs_get_uri_from_local_path (filename);
+  uri = g_filename_to_uri (filename, NULL, NULL);
   g_free (filename);
-  gnome_vfs_url_show (uri);
+ 
+  ret = cheese_open_uri (GTK_WINDOW (cheese_window->window), uri, &error);
+  if (ret == FALSE) {
+    dialog = gtk_message_dialog_new (GTK_WINDOW (cheese_window->window), 
+                                     GTK_DIALOG_DESTROY_WITH_PARENT, 
+	  	  		     GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, 
+				     _("Failed to launch program to show:\n%s"), uri);
+    gtk_dialog_run (GTK_DIALOG (dialog));
+    gtk_widget_destroy (dialog);
+    g_error_free(error);
+  }
   g_free (uri);
 }
 
@@ -193,7 +278,7 @@ cheese_window_cmd_save_as (GtkWidget *widget, CheeseWindow *cheese_window)
   filename = cheese_thumb_view_get_selected_image (CHEESE_THUMB_VIEW (cheese_window->thumb_view));
   g_return_if_fail (filename);
 
-  dialog = gtk_file_chooser_dialog_new ("Save File",
+  dialog = gtk_file_chooser_dialog_new (_("Save File"),
                                         GTK_WINDOW (cheese_window->window),
                                         GTK_FILE_CHOOSER_ACTION_SAVE,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
@@ -212,30 +297,25 @@ cheese_window_cmd_save_as (GtkWidget *widget, CheeseWindow *cheese_window)
   if (response == GTK_RESPONSE_ACCEPT)
   {
     char *target_filename;
+    GError *error = NULL;
+    gboolean ok;
 
     target_filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
+    GFile *target = g_file_new_for_path (target_filename);
 
-    char *source_uri = g_filename_to_uri (filename, NULL, NULL);
-    GnomeVFSURI *source = gnome_vfs_uri_new (source_uri);
-    g_free (source_uri);
+    GFile *source = g_file_new_for_path (filename);
 
-    char *target_uri = g_filename_to_uri (target_filename, NULL, NULL);
-    GnomeVFSURI *target = gnome_vfs_uri_new (target_uri);
-    g_free (target_uri);
+    ok = g_file_copy (source, target, G_FILE_COPY_OVERWRITE, NULL, NULL, NULL, &error);
 
-    response = gnome_vfs_xfer_uri (source, target,
-                                   GNOME_VFS_XFER_DEFAULT | GNOME_VFS_XFER_FOLLOW_LINKS,
-                                   GNOME_VFS_XFER_ERROR_MODE_ABORT,
-                                   GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
-                                   NULL, NULL);
-    gnome_vfs_uri_unref (source);
-    gnome_vfs_uri_unref (target);
+    g_object_unref (source);
+    g_object_unref (target);
 
-    if (response != GNOME_VFS_OK)
+    if (!ok)
     {
       char *header;
       GtkWidget *dlg;
 
+      g_error_free (error);
       header = g_strdup_printf (_("Could not save %s"), target_filename);
 
       dlg = gtk_message_dialog_new (GTK_WINDOW (cheese_window->window),
@@ -247,66 +327,39 @@ cheese_window_cmd_save_as (GtkWidget *widget, CheeseWindow *cheese_window)
       gtk_widget_destroy (dlg);
       g_free (header);
     }
+    g_free(target_filename);
   }
+  g_free(filename);
   gtk_widget_destroy (dialog);
 }
 
-
 static void
-cheese_window_cmd_move_file_to_trash (CheeseWindow *cheese_window, char *filename)
+cheese_window_cmd_move_file_to_trash (CheeseWindow *cheese_window, GList *files)
 {
-  GnomeVFSURI *uri;
-  GnomeVFSURI *trash_dir;
-  GnomeVFSURI *trash_uri;
-  int result;
-  char *name;
-  GError *error = NULL;
+  GError *error;
+  GList *l;
+  gchar *primary, *secondary;
+  GtkWidget *error_dialog;
 
-  uri = gnome_vfs_uri_new (g_filename_to_uri (filename, NULL, NULL));
-  result = gnome_vfs_find_directory (uri, GNOME_VFS_DIRECTORY_KIND_TRASH,
-                                     &trash_dir, FALSE, FALSE, 0777);
-  if (result != GNOME_VFS_OK)
-  {
-    gnome_vfs_uri_unref (uri);
-    char *header;
-    GtkWidget *dlg;
+  for (l = files; l != NULL; l = l->next) {
+    if (!g_file_trash (l->data, NULL, &error)) {
+      primary = g_strdup (_("Cannot move file to trash"));
+      secondary = g_strdup_printf (_("The file \"%s\" cannot be moved to the trash. Details: %s"),
+				   g_file_get_basename (l->data), error->message);
 
-    header = g_strdup_printf (_("Could not find the Trash"));
-    dlg = gtk_message_dialog_new (GTK_WINDOW (cheese_window->window),
-                                  GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-                                  GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, header);
-    gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dlg), error->message);
-    gtk_dialog_run (GTK_DIALOG (dlg));
-    gtk_widget_destroy (dlg);
-    g_free (header);
-    return;
-  }
+      error_dialog = gtk_message_dialog_new (GTK_WINDOW (cheese_window->window),
+ 					     GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+				             GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, primary);
+      gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (error_dialog),
+					        secondary);
+      gtk_dialog_run (GTK_DIALOG (error_dialog));
+      gtk_widget_destroy (error_dialog);
 
-  name = gnome_vfs_uri_extract_short_name (uri);
-  trash_uri = gnome_vfs_uri_append_file_name (trash_dir, name);
-  g_free (name);
-
-  result = gnome_vfs_move_uri (uri, trash_uri, TRUE);
-
-  gnome_vfs_uri_unref (uri);
-  gnome_vfs_uri_unref (trash_uri);
-  gnome_vfs_uri_unref (trash_dir);
-
-  if (result != GNOME_VFS_OK)
-  {
-    char *header;
-    GtkWidget *dlg;
-    char *basename = g_path_get_basename (filename);
-
-    header = g_strdup_printf (_("Error on deleting %s"), basename);
-    g_free (basename);
-    dlg = gtk_message_dialog_new (GTK_WINDOW (cheese_window->window),
-                                  GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-                                  GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, header);
-    gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dlg), error->message);
-    gtk_dialog_run (GTK_DIALOG (dlg));
-    gtk_widget_destroy (dlg);
-    g_free (header);
+      g_free (primary);
+      g_free (secondary);
+      g_error_free(error);
+      /*TODO if we can't move files to trash, maybe we should try to delete them....*/
+    }
   }
 }
 
@@ -317,19 +370,21 @@ cheese_window_move_all_media_to_trash (GtkWidget *widget, CheeseWindow *cheese_w
   char *prompt;
   int response;
   char *filename;
+  GFile *file;
+  GList *files_list = NULL;
   GDir *dir;
   char *path;
   const char *name;
 
-  prompt = g_strdup_printf (_("Are you sure you want to move\nall media to the trash?"));
+  prompt = g_strdup_printf (_("Really move all photos and videos to the trash?"));
   dlg = gtk_message_dialog_new_with_markup (GTK_WINDOW (cheese_window->window),
                                             GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-                                            GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
+                                            GTK_MESSAGE_WARNING, GTK_BUTTONS_NONE,
                                             "<span weight=\"bold\" size=\"larger\">%s</span>",
                                             prompt);
   g_free (prompt);
   gtk_dialog_add_button (GTK_DIALOG (dlg), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
-  gtk_dialog_add_button (GTK_DIALOG (dlg), _("Move to Trash"), GTK_RESPONSE_OK);
+  gtk_dialog_add_button (GTK_DIALOG (dlg), _("_Move to Trash"), GTK_RESPONSE_OK);
   gtk_dialog_set_default_response (GTK_DIALOG (dlg), GTK_RESPONSE_OK);
   gtk_window_set_title (GTK_WINDOW (dlg), "");
   gtk_widget_show_all (dlg);
@@ -347,46 +402,34 @@ cheese_window_move_all_media_to_trash (GtkWidget *widget, CheeseWindow *cheese_w
     if (g_str_has_suffix (name, PHOTO_NAME_SUFFIX) || 
         g_str_has_suffix (name, VIDEO_NAME_SUFFIX))
     {
-      filename = g_strjoin ("/", path, name, NULL);
-      cheese_window_cmd_move_file_to_trash (cheese_window, filename);
+      filename = g_strjoin (G_DIR_SEPARATOR_S, path, name, NULL);
+      file = g_file_new_for_path (filename);
+
+      files_list = g_list_append (files_list, file);
       g_free (filename);
     }
   }
+  cheese_window_cmd_move_file_to_trash (cheese_window, files_list);
+  g_list_free (files_list);
+  g_dir_close(dir);
 }
 
 static void
 cheese_window_move_media_to_trash (GtkWidget *widget, CheeseWindow *cheese_window)
 {
-  GtkWidget *dlg;
-  char *prompt;
-  int response;
-  char *filename, *basename;
+  char *filename;
+  GFile *file;
+  GList *files_list = NULL;
 
   filename = cheese_thumb_view_get_selected_image (CHEESE_THUMB_VIEW (cheese_window->thumb_view));
   g_return_if_fail (filename);
 
-  basename = g_path_get_basename (filename);
-  prompt = g_strdup_printf (_("Are you sure you want to move\n\"%s\" to the trash?"), basename);
-  g_free (basename);
-  dlg = gtk_message_dialog_new_with_markup (GTK_WINDOW (cheese_window->window),
-                                            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-                                            GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
-                                            "<span weight=\"bold\" size=\"larger\">%s</span>",
-                                            prompt);
-  g_free (prompt);
-  gtk_dialog_add_button (GTK_DIALOG (dlg), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
-  gtk_dialog_add_button (GTK_DIALOG (dlg), _("Move to Trash"), GTK_RESPONSE_OK);
-  gtk_dialog_set_default_response (GTK_DIALOG (dlg), GTK_RESPONSE_OK);
-  gtk_window_set_title (GTK_WINDOW (dlg), "");
-  gtk_widget_show_all (dlg);
+  file = g_file_new_for_path (filename);
+  g_free(filename);
 
-  response = gtk_dialog_run (GTK_DIALOG (dlg));
-  gtk_widget_destroy (dlg);
-
-  if (response !=  GTK_RESPONSE_OK)
-    return;
-
-  cheese_window_cmd_move_file_to_trash (cheese_window, filename);
+  files_list = g_list_append (files_list, file);
+  cheese_window_cmd_move_file_to_trash (cheese_window, files_list);
+  g_list_free (files_list);
 }
 
 static void
@@ -437,13 +480,16 @@ cheese_window_cmd_set_about_me_photo (GtkWidget *widget, CheeseWindow *cheese_wi
         gtk_dialog_run (GTK_DIALOG (dlg));
         gtk_widget_destroy (dlg);
         g_free (header);
+        g_error_free (error);
       }
       g_free (*data);
     }
+    g_free(filename);
     g_object_unref (pixbuf);
   }
 }
 
+
 static void
 cheese_window_cmd_command_line (GtkAction *action, CheeseWindow *cheese_window)
 {
@@ -491,50 +537,21 @@ static void
 cheese_window_cmd_help_contents (GtkAction *action, CheeseWindow *cheese_window)
 {
   GError *error = NULL;
-  char *command;
-  const char *lang;
-  char *uri = NULL;
-  GdkScreen *gscreen;
-  // TODO what do we want for link_id
-  char *link_id = NULL;
+  gboolean ret;
 
-  int i;
+  ret = cheese_open_uri (GTK_WINDOW (cheese_window->window), "ghelp:cheese", &error);
 
-  const char * const * langs = g_get_language_names ();
-
-  for (i = 0; langs[i]; i++) 
-  {
-    lang = langs[i];
-    if (strchr (lang, '.')) 
-      continue;
-
-    uri = g_build_filename(DATADIR, "/gnome/help/cheese/", lang, "/cheese.xml", NULL);
-
-    if (g_file_test (uri, G_FILE_TEST_EXISTS))
-      break;
-  }
-
-  if (link_id)
-    command = g_strconcat ("gnome-open ghelp://", uri, "?", link_id, NULL);
-  else
-    command = g_strconcat ("gnome-open ghelp://", uri,  NULL);
-
-
-  gscreen = gdk_screen_get_default();
-  gdk_spawn_command_line_on_screen (gscreen, command, &error);
-  if (error != NULL) 
+  if (ret == FALSE) 
   {
     GtkWidget *d;
-
-    d = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-                               GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, error->message);
+    d = gtk_message_dialog_new (GTK_WINDOW (cheese_window->window), 
+                                GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+                                GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, 
+                                _("Unable to open help file for Cheese"));
     gtk_dialog_run(GTK_DIALOG(d));
     gtk_widget_destroy(d);
     g_error_free(error);
   }
-
-  g_free (command);
-  g_free (uri);
 }
 
 static void
@@ -751,6 +768,74 @@ cheese_window_photo_video_toggle_buttons_cb (GtkWidget *widget, CheeseWindow *ch
 
 }
 
+void
+cheese_window_countdown_hide_cb (gpointer data)
+{
+  CheeseWindow *cheese_window = (CheeseWindow *) data;
+  gtk_notebook_set_current_page(GTK_NOTEBOOK(cheese_window->notebook_bar), 0);
+}
+
+void
+cheese_window_countdown_picture_cb (gpointer data)
+{
+  CheeseWindow *cheese_window = (CheeseWindow *) data;
+  GError *error = NULL;
+  GstAudioPlay *audio_play;
+  char *file;
+
+  file = audio_play_get_filename (cheese_window);
+  audio_play = gst_audio_play_file (file, &error);
+  if (!audio_play) 
+  {
+    g_warning (error ? error->message : "Unknown error");
+    g_error_free (error);
+  }
+
+  g_free (file);
+
+  cheese_window->photo_filename = cheese_fileutil_get_new_media_filename (WEBCAM_MODE_PHOTO);
+  cheese_webcam_take_photo (cheese_window->webcam, cheese_window->photo_filename);
+}
+
+static void
+cheese_window_stop_recording (CheeseWindow *cheese_window)
+{
+  if (cheese_window->recording)
+  {
+    cheese_webcam_stop_video_recording (cheese_window->webcam);
+    gtk_widget_set_sensitive (cheese_window->button_effects, TRUE);
+    gtk_widget_set_sensitive (cheese_window->button_photo, TRUE);
+    gtk_widget_set_sensitive (cheese_window->button_video, TRUE);
+    gchar * str = g_strconcat ("<b>", _("_Start recording"), "</b>", NULL);
+    gtk_label_set_text_with_mnemonic (GTK_LABEL (cheese_window->label_take_photo), str);
+    g_free (str);
+    gtk_label_set_use_markup (GTK_LABEL (cheese_window->label_take_photo), TRUE);
+    gtk_image_set_from_stock (GTK_IMAGE (cheese_window->image_take_photo), GTK_STOCK_MEDIA_RECORD, GTK_ICON_SIZE_BUTTON);
+    
+    cheese_window->recording = FALSE;
+  }
+}
+
+static gboolean
+cheese_window_cancel_cb (CheeseWindow *cheese_window,
+			 GtkAccelGroup *accel_group,
+			 guint keyval, GdkModifierType modifier)
+{
+  cheese_countdown_cancel ((CheeseCountdown *) cheese_window->countdown);
+  
+  if (cheese_window->webcam_mode == WEBCAM_MODE_PHOTO)
+  {
+    gtk_notebook_set_current_page (GTK_NOTEBOOK (cheese_window->notebook_bar), 0);
+      
+    gtk_widget_set_sensitive (cheese_window->take_picture, TRUE);
+  }
+  else
+  {
+    cheese_window_stop_recording (cheese_window);
+  }
+  return TRUE;
+}
+
 static void
 cheese_window_action_button_clicked_cb (GtkWidget *widget, CheeseWindow *cheese_window)
 {
@@ -758,11 +843,11 @@ cheese_window_action_button_clicked_cb (GtkWidget *widget, CheeseWindow *cheese_
 
   if (cheese_window->webcam_mode == WEBCAM_MODE_PHOTO)
   {
-    gtk_widget_set_sensitive (cheese_window->take_picture, FALSE);
-    cheese_audio_play_file (cheese_window->audio_play);
+    cheese_countdown_start((CheeseCountdown *) cheese_window->countdown, cheese_window_countdown_picture_cb, cheese_window_countdown_hide_cb, (gpointer) cheese_window);
+    gtk_notebook_set_current_page (GTK_NOTEBOOK(cheese_window->notebook_bar), 1);
 
-    cheese_window->photo_filename = cheese_fileutil_get_new_media_filename (WEBCAM_MODE_PHOTO);
-    cheese_webcam_take_photo (cheese_window->webcam, cheese_window->photo_filename);
+    gtk_widget_set_sensitive (cheese_window->take_picture, FALSE);
+    // FIXME: set menu inactive
   }
   else if (cheese_window->webcam_mode == WEBCAM_MODE_VIDEO)
   {
@@ -779,20 +864,13 @@ cheese_window_action_button_clicked_cb (GtkWidget *widget, CheeseWindow *cheese_
 
       cheese_window->video_filename = cheese_fileutil_get_new_media_filename (WEBCAM_MODE_VIDEO);
       cheese_webcam_start_video_recording (cheese_window->webcam, cheese_window->video_filename);
+
+      cheese_window->recording = TRUE;
     }
     else
     {
-      cheese_webcam_stop_video_recording (cheese_window->webcam);
-      gtk_widget_set_sensitive (cheese_window->button_effects, TRUE);
-      gtk_widget_set_sensitive (cheese_window->button_photo, TRUE);
-      gtk_widget_set_sensitive (cheese_window->button_video, TRUE);
-      str = g_strconcat ("<b>", _("_Start recording"), "</b>", NULL);
-      gtk_label_set_text_with_mnemonic (GTK_LABEL (cheese_window->label_take_photo), str);
-      g_free (str);
-      gtk_label_set_use_markup (GTK_LABEL (cheese_window->label_take_photo), TRUE);
-      gtk_image_set_from_stock (GTK_IMAGE (cheese_window->image_take_photo), GTK_STOCK_MEDIA_RECORD, GTK_ICON_SIZE_BUTTON);
+      cheese_window_stop_recording (cheese_window);
     }
-    cheese_window->recording = !cheese_window->recording;
   }
 }
 
@@ -928,6 +1006,7 @@ cheese_window_create_window (CheeseWindow *cheese_window)
   cheese_window->label_video         = glade_xml_get_widget (gxml, "label_video");
   cheese_window->main_vbox           = glade_xml_get_widget (gxml, "main_vbox");
   cheese_window->notebook            = glade_xml_get_widget (gxml, "notebook");
+  cheese_window->notebook_bar        = glade_xml_get_widget (gxml, "notebook_bar");
   cheese_window->screen              = glade_xml_get_widget (gxml, "video_screen");
   cheese_window->take_picture        = glade_xml_get_widget (gxml, "take_picture");
 
@@ -941,9 +1020,7 @@ cheese_window_create_window (CheeseWindow *cheese_window)
   gtk_container_add (GTK_CONTAINER (cheese_window->thumb_scrollwindow), cheese_window->thumb_view);
 
   char *gconf_effects;
-  g_object_get (cheese_window->gconf,
-               "gconf_prop_selected_effects", &gconf_effects,
-               NULL);
+  g_object_get (cheese_window->gconf, "gconf_prop_selected_effects", &gconf_effects, NULL);
   cheese_window->effect_frame        = glade_xml_get_widget (gxml, "effect_frame");
   cheese_window->effect_chooser      = cheese_effect_chooser_new (gconf_effects);
   gtk_container_add (GTK_CONTAINER (cheese_window->effect_frame), cheese_window->effect_chooser);
@@ -955,6 +1032,11 @@ cheese_window_create_window (CheeseWindow *cheese_window)
   gtk_container_add (GTK_CONTAINER (cheese_window->throbber_frame), cheese_window->throbber);
   gtk_widget_show (cheese_window->throbber);
 
+  cheese_window->countdown_frame      = glade_xml_get_widget (gxml, "countdown_frame");
+  cheese_window->countdown            = cheese_countdown_new ();
+  gtk_container_add (GTK_CONTAINER (cheese_window->countdown_frame), cheese_window->countdown);
+  gtk_widget_show (cheese_window->countdown);
+
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (cheese_window->button_photo), TRUE);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (cheese_window->button_video), FALSE);
 
@@ -1037,6 +1119,12 @@ cheese_window_create_window (CheeseWindow *cheese_window)
 
   gtk_window_add_accel_group (GTK_WINDOW (cheese_window->window), 
                               gtk_ui_manager_get_accel_group (cheese_window->ui_manager));
+  gtk_accel_group_connect (gtk_ui_manager_get_accel_group (cheese_window->ui_manager),
+			   GDK_Escape, 0, 0,
+			   g_cclosure_new_swap (G_CALLBACK (cheese_window_cancel_cb),
+						cheese_window, NULL));
+
+
   gtk_action_group_set_sensitive (cheese_window->actions_file, FALSE);
 
   /* Default handlers for closing the application */
@@ -1052,19 +1140,42 @@ cheese_window_create_window (CheeseWindow *cheese_window)
                     "clicked", G_CALLBACK (cheese_window_effect_button_pressed_cb), cheese_window);
 
   g_signal_connect (cheese_window->thumb_view, "button_press_event",
-                          G_CALLBACK (cheese_window_button_press_event_cb), cheese_window);
+                    G_CALLBACK (cheese_window_button_press_event_cb), cheese_window);
+}
+
+void 
+setup_camera (CheeseWindow *cheese_window)
+{
+  char *webcam_device = NULL;
+
+  g_object_get (cheese_window->gconf, "gconf_prop_webcam", &webcam_device, NULL);
+
+  cheese_window->webcam = cheese_webcam_new (cheese_window->screen, webcam_device);
+  g_free (webcam_device);
+
+  g_signal_connect (cheese_window->webcam, "photo-saved",
+                    G_CALLBACK (cheese_window_photo_saved_cb), cheese_window);
+  g_signal_connect (cheese_window->webcam, "video-saved",
+                    G_CALLBACK (cheese_window_video_saved_cb), cheese_window);
+
+  cheese_webcam_set_effect (cheese_window->webcam, 
+                            cheese_effect_chooser_get_selection (CHEESE_EFFECT_CHOOSER (cheese_window->effect_chooser)));
+
+  cheese_webcam_play (cheese_window->webcam);
+  gtk_notebook_set_current_page (GTK_NOTEBOOK(cheese_window->notebook), 0);
+  ephy_spinner_stop (EPHY_SPINNER (cheese_window->throbber));
 }
 
 void
 cheese_window_init ()
 {
   CheeseWindow *cheese_window;
-  char *webcam_device = NULL;
 
   cheese_window = g_new (CheeseWindow, 1);
 
   cheese_window->gconf = cheese_gconf_new ();
-  cheese_window->audio_play = cheese_audio_play_new ();
+  cheese_window->audio_play_counter = 0;
+  cheese_window->rand = g_rand_new ();
 
   cheese_window_create_window (cheese_window);
  
@@ -1076,25 +1187,15 @@ cheese_window_init ()
   cheese_window->webcam_mode = WEBCAM_MODE_PHOTO;
   cheese_window->recording = FALSE;
   
-  g_object_get (cheese_window->gconf, "gconf_prop_webcam", &webcam_device, NULL);
-
-  cheese_window->webcam = cheese_webcam_new (cheese_window->screen, webcam_device);
-  g_free (webcam_device);
-
-  g_signal_connect (cheese_window->webcam, "photo-saved",
-                    G_CALLBACK (cheese_window_photo_saved_cb), cheese_window);
-  g_signal_connect (cheese_window->webcam, "video-saved",
-                    G_CALLBACK (cheese_window_video_saved_cb), cheese_window);
+  /* Run cam setup in its own thread */
+  GError *error = NULL;
+  if (!g_thread_create ((GThreadFunc) setup_camera, cheese_window, FALSE, &error)) {
+    g_error ("Failed to create setup thread: %s\n", error->message);
+    g_error_free (error);
+    return;
+  }
 
   /* Make URLs and email clickable in about dialog */
   gtk_about_dialog_set_url_hook (cheese_about_dialog_handle_url, NULL, NULL);
   gtk_about_dialog_set_email_hook (cheese_about_dialog_handle_email, NULL, NULL);
-
-  cheese_webcam_set_effect (cheese_window->webcam, 
-                            cheese_effect_chooser_get_selection (CHEESE_EFFECT_CHOOSER (cheese_window->effect_chooser)));
-
-  cheese_webcam_play (cheese_window->webcam);
-  gtk_notebook_set_current_page (GTK_NOTEBOOK(cheese_window->notebook), 0);
-  ephy_spinner_stop (EPHY_SPINNER (cheese_window->throbber));
 }
-
